-- Project: GridConfigurableLayouts 0.5.0.5
-- File: LDBQuickSelect.lua
-- Last Modified: 2012-09-28T09:02:09Z
-- Author: msaint
-- Project desc: Allows in-game creation of highly customized layouts in Grid
-- File desc: Allows quick selection of layout from Grid LDB icon using button4 

-- Globals to ignore:
-- GLOBALS: DEFAULT_CHAT_FRAME, UIParent, RaidNotice_AddMessage, RaidWarningFrame, ChatTypeInfo, GameTooltip

local DEBUG = false
local debug = DEBUG and function(s) DEFAULT_CHAT_FRAME:AddMessage("GCL_LDBQS: "..s, 1, 0, 0) end or function() return end  

local FILE_VERSION = tonumber("98") or 5000

--{{{ Get or Create the module
local gcl = Grid:GetModule("GridConfigurableLayouts", true) or Grid:NewModule("GridConfigurableLayouts", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
gcl.Versions = gcl.Versions or {} 
if (gcl.Versions.LDBQuickSelect and gcl.Versions.LDBQuickSelect >= FILE_VERSION) then return end
gcl.Versions.LDBQuickSelect = FILE_VERSION
---}}}


--{{{ Libraries and modules
--local ldb = LibStub:GetLibrary("LibDataBroker-1.1", true)
local gridLDB = (LibStub:GetLibrary("LibDataBroker-1.1")):GetDataObjectByName("Grid")
local GridRoster = Grid:GetModule("GridRoster")
local GridLayout = Grid:GetModule("GridLayout")
local Dewdrop = LibStub("Dewdrop-2.0")
--}}}


--{{{ Hack the localization function for easy implementation later
gcl.L = gcl.L or {} 
local L = gcl.L 
setmetatable(L,{ __index = function(l, text) return text end})
---}}}


---{{{ Make everthing local that we can
   --Lua library:
local next, ipairs, pairs, table, wipe, string, tostring, type 
       = next, ipairs, pairs, table, wipe, string, tostring, type  
   --WoW api:
local InCombatLockdown = InCombatLockdown
---}}}


--{{{ make debug function local for convenience
-- local function debug(...)
--    gcl:Debug("[GCL-LDB]", ...)
-- end
--}}}


--{{{ Local variables and objects
local gclDB = {}
local gridLayoutList = {}
local gclLDBOptions = {type = "group", args = {}}

local gclLDBOptionsArgsBase = {
   gridLock = {
		type = "toggle",
      order = 10,
      name = L["Gridframes Locked"],
      desc = L["Lock or unlock the position of the Grid frames."],
		get = function() return GridLayout.db.profile.FrameLock end,
		set = function() GridLayout.db.profile.FrameLock = not GridLayout.db.profile.FrameLock end,
		
	},
   gridLayouts = {
      type = "group",
      order = 20,
      name = L["Other Layouts"],
      desc = L["Show all Grid layouts."],
      args = gridLayoutList,
      
   },   
   gclLayoutsHeader = {
      type = "header",
      order = 99,
      name = L["GCL Layouts"],
   },
}


local function getCurrentLayout()
   return GridLayout.db.profile.layouts[GridRoster:GetPartyState()]   
end

local function setCurrentLayout(layoutName)
   local partyState = GridRoster:GetPartyState() 
   if GridLayout.db.profile.layouts[partyState] ~= layoutName 
         and GridLayout.layoutSettings[layoutName] then 
      GridLayout.db.profile.layouts[partyState] = layoutName
      GridLayout:ReloadLayout()
      RaidNotice_AddMessage(RaidWarningFrame, string.format(L["Layout for %q changed to %q"], partyState, layoutName), ChatTypeInfo["RAID_WARNING"])
   end  
end

local function addToggle(argTable, id, name, order, get, set, desc)
   if argTable and (type(argTable) == "table") and id and name 
          and get and (type(get) == "function") 
         and set and (type(set) == "function") then
      desc = desc or name
      argTable[id] = {type = "toggle", pass = true, passValue = name, name = name, order = order, get = get, set = set, desc = desc}
      return true
   end
end

local function addLayoutToggles(argTable, sourceList, excludeList, startOrderNum)
   if argTable and (type(argTable) == "table") 
         and sourceList and (type(sourceList) == "table")
         and ((not excludeList) or (type(excludeList) == "table")) then 
      startOrderNum = startOrderNum or 1
      local togList = {}
      local function get(v)
         return (v == getCurrentLayout())
      end
      local function set(v)
         setCurrentLayout(v)
      end 
      for loName, _ in pairs(sourceList) do
         if not (excludeList and excludeList[loName]) then
            table.insert(togList, loName)
         end 
      end
      if next(togList) then
         table.sort(togList)
      end
      for i, loName in ipairs(togList) do
         addToggle(argTable, loName.."Toggle"..tostring(i), loName, startOrderNum + i, get, set)
      end
   end
end

-- Hook the existing click and tooltip handlers on the Grid icon.
-- These must be hooked manually because of peculiarities in the way acehook
-- and LDB work.
local oldOnClick = gridLDB["OnClick"]
gridLDB["OnClick"] = function (self, button)
   if (button == "LeftButton") and gclDB.ldbQuickSelect and not InCombatLockdown() then
      GameTooltip:Hide()
      -- Add gcl layouts to the first level dropdown
      wipe(gclLDBOptions.args)
      for k, v in pairs(gclLDBOptionsArgsBase) do
         gclLDBOptions.args[k] = v
      end
      addLayoutToggles(gclLDBOptions.args, gclDB.layouts, nil, 100)
      -- Add all other layouts to the level two dropdown
      wipe(gridLayoutList)
      addLayoutToggles(gridLayoutList, GridLayout.layoutSettings, gclDB.layouts)
      -- Open the dropdown
      Dewdrop:Open(UIParent, "children", gclLDBOptions, "cursorX", true, "cursorY", true)
   else
      oldOnClick(self, button)
	end
end

local oldOnTooltipShow = gridLDB["OnTooltipShow"]
gridLDB["OnTooltipShow"] = function(tooltip)
	tooltip:AddLine("Grid", 1, 1, 1)
	if gclDB.ldbQuickSelect then
      if InCombatLockdown() then 
      	tooltip:AddLine(L["Click to change layouts."], 0.5, 0.5, 0.5)
      	tooltip:AddLine(L["Right-Click for more options."], 0.5, 0.5, 0.5)
		else
			tooltip:AddLine(L["Click to change layouts."])
			tooltip:AddLine(L["Right-Click for more options."])
		end
	else
	   oldOnTooltipShow(tooltip)
	end
end

Grid.options.args.quickselect = {
	order = -2,
	name = L["Use minimap icon to change layouts."],
	desc = L["Right clicking the Grid minimap or LDB icon brings up a layout quickselect menu but does not toggle Grid frame lock."],
	width = "double",
	type = "toggle",
  	get = function () return gclDB.ldbQuickSelect end,
  	set = function () gclDB.ldbQuickSelect = not gclDB.ldbQuickSelect end,
}


function gcl:initLDBQuickSelect()
   gclDB = gcl.gcldb
end
