-- LibSpecRoster
-- Project: LibSpecRoster 0.5.0.5
-- File: LibSpecRoster.lua
-- Last Modified: 2012-09-29T06:11:21Z
-- Author: msaint
-- Project desc: Efficiently maintains up to date raid spec information, inspecting
--               only when needed. 


--  Library methods:
--    lib.RegisterMessage(yourObject, "messageName", yourMethod)
--    lib.UnregisterMessage(yourObject, "messageName")
--    lib.RegisterAllMessages(yourObject)
--    lib:getUnitSpecID(unit) - returns specialization ID
--    lib:getGUIDSpecID(guid) - returns specialization ID
--    lib:getUnitRole(unit) - returns role ("tank" | "healer" | "melee" | "ranged")
--    lib:getGUIDROle(guid) - returns role
--    lib:getUnitSpecInfo(unit) - returns specID, specName, class, role, blizzard role ("TANK" | "HEALER" | "DAMAGER") 
--    lib:getGUIDSpecInfo(guid)
--
--  Messages:
--    "LSR_SpecializationChanged" - params: eventName, unit, specID, role
-- 
--  To do: add "LSR_RoleChanged" ?

-- Globals to ignore:
-- GLOBALS: DEFAULT_CHAT_FRAME

local DEBUG = false
local debug = DEBUG and function(s) DEFAULT_CHAT_FRAME:AddMessage("LSR: "..s, 1, 0, 0) end or function() return end  


-- ***** Obtain the library object **********
local LIB_VERSION = tonumber("100") or 5000
local MAJOR, MINOR = "LibSpecRoster-1.0", LIB_VERSION
local lib = LibStub:NewLibrary(MAJOR, MINOR)
if not lib then return end
local msg = LibStub("CallbackHandler-1.0"):New(lib, "RegisterMessage", "UnregisterMessage", "UnregisterAllMessages")


-- ***** LUA library functions made local **********
local math, max
      = math, math.max
local table, tinsert, tremove
      = table, tinsert, tremove
local pairs, ipairs, next, tostring, select
      = pairs, ipairs, next, tostring, select
local s = tostring


-- ***** API functions made local **********
local GetSpellInfo, UnitIsVisible, GetNumGroupMembers, GetPlayerInfoByGUID, IsInRaid, UnitGUID, UnitClass 
      = GetSpellInfo, UnitIsVisible, GetNumGroupMembers, GetPlayerInfoByGUID, IsInRaid, UnitGUID, UnitClass
local UnitIsConnected, CanInspect, UnitIsUnit
      = UnitIsConnected, CanInspect, UnitIsUnit
local GetSpecializationInfo, GetSpecialization, GetSpecializationInfoByID, GetInspectSpecialization
      = GetSpecializationInfo, GetSpecialization, GetSpecializationInfoByID, GetInspectSpecialization
local RegisterEvent, UnregisterEvent, UnregisterAllEvents
      = RegisterEvent, UnregisterEvent, UnregisterAllEvents      


-- ***** Constants-ish **********
-- time delays
local updateDelay = 0.120 -- miliseconds between updates / talent requests

-- Unit statuses
local GOOD, INVALID, QUEUED, REQUEUE, REQUESTED, REQUESTED_GOOD = "GOOD", "INVALID", "QUEUED", "REQUEUE", "REQUESTED", "REQUESTED_GOOD"

-- Role constants
local fixedRole = {
   ROGUE = "melee",
   HUNTER = "ranged",
   MAGE = "ranged",
   WARLOCK = "ranged",
}
local defaultDamageRole = {
   WARRIOR = "melee",
   DEATHKNIGHT = "melee",
   PALADIN = "melee",
   MONK = "melee",
}
setmetatable(defaultDamageRole,{ __index = function(l, class) return fixedRole[class] or "ranged" end})
local specIDRoles = {
      -- Death Knight
      [250] = "tank", -- Blood
      [251] = "melee", -- Frost
      [252] = "melee", -- Unholy
      -- Druid
      [102] = "ranged", -- Balance
      [103] = "melee", -- Feral
      [104] = "tank", -- Guardian
      [105] = "healer", -- Restoration
      -- Hunter
      [253] = "ranged", -- Beast Mastery
      [254] = "ranged", -- Marksmanship
      [255] = "ranged", -- Survival
      -- Mage
      [62] = "ranged", -- Arcane
      [63] = "ranged", -- Fire
      [64] = "ranged", -- Frost
      -- Monk
      [268] = "tank", -- Brewmaster
      [269] = "healer", -- Mistweaver
      [270] = "melee", -- Windwalker
      -- Paladin
      [65] = "healer", -- Holy
      [66] = "tank", -- Protection
      [67] = "melee", -- Retribution
      -- Priest
      [256] = "healer", -- Discipline
      [257] = "healer", -- Holy
      [258] = "ranged", -- Shadow
      -- Rogue
      [259] = "melee", -- Assassination
      [260] = "melee", -- Combat
      [261] = "melee", -- Subtlety
      -- Shaman
      [262] = "ranged", -- Elemental
      [263] = "melee", -- Enhancement
      [264] = "healer", -- Restoration
      -- Warlock
      [265] = "ranged", -- Affliction
      [266] = "ranged", -- Demonology
      [267] = "ranged", -- Destruction
      -- Warrior
      [71] = "melee", -- Arms
      [72] = "melee", -- Fury
      [73] = "tank", -- Protection
   }


-- ***** Local tables **********
local events = {} -- Used by event handlers and timers
local group = {} -- All raid members: group[guid] = {unitID, specID, specName, class, status} 


-- ***** Local variables **********
local playerGUID -- Set in init() and used in setStatus()
--local houston = false -- if true, we are having problems getting class or guid info from the server


-- ***** Event Frame Setup **********
local eFrame = _G["LibSpecRosterEventFrame"] or CreateFrame("Frame", "LibSpecRosterEventFrame", UIParent)
eFrame:SetScript("OnEvent", function(self, event, ...)
      if events[event] then
         events[event](events, ...)
      end
   end) 
eFrame:RegisterEvent("PLAYER_ENTERING_WORLD") -- This is where we will initialize the library


-- ***** Local utility functions **********
local function uniter() 
-- use is: for i, unitID in uniter() do block end
-- i starts at 1 and corresponds to raid or party unit number
-- for party (and solo), player is 'player' and is given last 
   local raid, N = IsInRaid(), GetNumGroupMembers() 
   N = N + ((N==0 or raid) and 1 or 0) 
   local iter = function(prefix, i)
      i = i + 1
      if (i < N) then
         return i, prefix..tostring(i)
      elseif (i == N and not raid) then
         return i, "player"
      else 
         return nil
      end
   end           
   return iter, (raid and "raid" or "party"), 0
end

local function buildRoster()
   local roster, guidRoster, guid = {}, {}
   for i, unit in uniter() do
      guid = UnitGUID(unit)
      if guid then
         roster[guid] = unit
      else
         debug("Houston, we need a procedure.  GUID was not available for "..s(unit))
      end
   end
   return roster
end

local setStatus
do
   local lock 
   setStatus = function (guid, status)
      if not (guid and status) then return end
      while lock do end -- This is not perfect, but also not as bad as it looks
      lock = true
      local rec, doFire = group[guid]
      if rec then
         local recStatus = rec.status
         if (status == GOOD) then
            if (recStatus == REQUESTED) or (recStatus == REQUESTED_GOOD) or (guid == playerGUID) then
               -- 'player' info is instantly available, so if you set it to GOOD, I believe you.
               rec.status = GOOD
               doFire = true
            end
         elseif (status == INVALID) then
            if (recStatus == GOOD) then
               rec.status = INVALID
            end
         elseif (status == REQUESTED) then
            rec.status = (recStatus == GOOD) and REQUESTED_GOOD or REQUESTED
         else
            rec.status = status
         end
      end
      lock = nil
      debug("Status being set to "..s(status).." ...")
      if rec then
         debug(" ... but really is "..s(rec.status))
      end
      if doFire then 
         msg:Fire("LSR_SpecializationChanged", rec.unitID, guid, rec.specID, specIDRoles[rec.specID])
      end
   end
end

local function getRole(guid)
   local rec, role, isDefinite = group[guid]
   if rec then
      local class = rec.class or select(2, UnitClass(rec.unitID)) or "CLASSLESS"
      role = (rec.specID and (rec.specID > 0) and specIDRoles[rec.specID]) or defaultDamageRole[class]
      isDefinite = (fixedRole[class] or (rec.status == GOOD)) and true
      return role, isDefinite
   end
end


-- ***** OnUpdate - Where it all happens! **********
local OnUpdate
do
   local sinceLast, queue = 0, {}, {}
   local checkOnce = true
   OnUpdate = function (self, elapsed)
      sinceLast = sinceLast + elapsed
      if (sinceLast >= updateDelay) then
         sinceLast = updateDelay - sinceLast -- REALLY do it every updateDelay ms
         
         if checkOnce then
            checkOnce = nil
         end
         
         for guid, record in pairs(group) do
            if (record.status == INVALID) or (record.status == REQUEUE) then
               tinsert(queue, 1, guid)
               setStatus(guid, QUEUED)
            end
         end
         
         for i = #queue, 1, -1 do
            local guid = queue[i] 
            local rec = group[guid]
            local unit = rec.unitID
            if (rec.status == QUEUED) and unit and CanInspect(unit) then
               if UnitIsConnected(unit) then
                  debug("Sending NotifyInspect for "..s(unit))
                  NotifyInspect(unit) -- REQUESTED status is set by our secure hook on NotifyInspect
                  tremove(queue, i)
                  break -- Only one request per requestThrottleDelay!
               elseif CanInspect(unit) then -- Always true for group members, even if they are offline
                  setStatus(guid, REQUEUE)
               end
            end
            tremove(queue, i)
         end
      end            
   end
end


-- ***** Library initialization **********
local function init()
   debug("init was reached")
   playerGUID = UnitGUID("player") -- Used by setStatus
   -- Create initial roster before handling events
   events:GROUP_ROSTER_UPDATE()               
   eFrame:UnregisterAllEvents()
   for e, _ in pairs(events) do
      eFrame:RegisterEvent(e)
   end
   eFrame:Show()
end


function events:PLAYER_ENTERING_WORLD()
   debug("PLAYER_ENTERING_WORLD event handler was reached.")
   -- Ridiculous hack here just because I want to index by GUID
   local sinceLast, delay = 0, 0.150

   local function doWaitForGUIDs(self, elapsed)
      sinceLast = sinceLast + elapsed
      if sinceLast >= delay then
         for _, unit in uniter() do
            if not UnitGUID(unit) then
               sinceLast = 0
               return
            end
         end
         eFrame:Hide()
         eFrame:SetScript("OnUpdate", OnUpdate)
         init()
      end 
   end
   eFrame:SetScript("OnUpdate", doWaitForGUIDs)
   eFrame:Show()
   events.PLAYER_ENTERING_WORLD = nil
end


-- ***** Event handlers **********
function events:GROUP_ROSTER_UPDATE()
   local newRoster = buildRoster()
   for guid, rec in pairs(group) do -- Discard or update existing entries in group
      if not newRoster[guid] then -- discard
         group[guid] = nil
      else -- same toon, new unitID: update and remove from newRoster
         rec.unitID = newRoster[guid]
         newRoster[guid] = nil
      end
   end 
   for guid, unit in pairs(newRoster) do -- Any units still in newRoster must be added to group 
      debug("Adding "..s(unit).." to the roster.")
      group[guid] = {unitID = unit, status = INVALID}
      group[guid].class = UnitClass(unit)
      if UnitIsUnit(unit, "player") then
         -- We should get this now: it shouldn't end up queued
         events:PLAYER_SPECIALIZATION_CHANGED("player")
      end
   end
end      


function events:PLAYER_SPECIALIZATION_CHANGED(unit)
   unit = unit or "player" -- This event sometimes leaves out unit for the player 
   debug("PLAYER_SPECIALIZATION_CHANGED event for "..s(unit))
   local guid = UnitGUID(unit)
   if guid then
      local record, newStatus = group[guid], INVALID
      if UnitIsUnit(unit, "player") then
         local spec, specName, _, _, _, blizRole = GetSpecializationInfo((GetSpecialization())) 
         if spec and spec > 0 then
            record.specID, record.specName, record.blizRole = spec, specName, blizRole 
            record.class = record.class or UnitClass(unit)
            newStatus = GOOD            
         end
      end
      setStatus(guid, newStatus)
   else
      debug("GUID was not returned for "..unit)
   end            
end


function events:INSPECT_READY(guid)
   debug("INSPECT_READY received for "..s(guid))
   local record = guid and group[guid]
   if record then
      local spec, specName, _, _, _, blizRole, class = GetSpecializationInfoByID(GetInspectSpecialization(record.unitID))
      if spec and spec > 0 then
         -- Even if the status doesn't become GOOD, this is the best data we have, so use it.
         record.specID, record.specName, record.blizRole, record.class = spec, specName, blizRole, class
         setStatus(guid, GOOD)
      end
   end
end


-- ***** Hook to catch inspects from other addons and avoid duplicating them **********
local function notifyInspect(unit)
   local guid = UnitGUID(unit)
   if (guid and group[guid] and CanInspect(unit) and UnitIsConnected(unit)) then
	   -- If we know the request is going to fail, don't flag it as REQUESTED
      setStatus(guid, REQUESTED)
	end
end
hooksecurefunc("NotifyInspect", notifyInspect)


-- ***** Exposed Library Methods **********
function lib:getUnitSpecID(unit)
   local guid = unit and UnitGUID(unit)
   local rec = guid and group[guid]
   if rec then
      return rec.specID, ((rec.status == GOOD) or (rec.status == REQUESTED_GOOD))
   end
end

function lib:getGUIDSpecID(guid)
   local rec = guid and group[guid]
   if rec then
      return rec.specID, ((rec.status == GOOD) or (rec.status == REQUESTED_GOOD))
   end
end

function lib:getUnitRole(unit)
   local guid = UnitGUID(unit)
   if guid and group[guid] then
      return getRole(guid)
   end
end

function lib:getGUIDROle(guid)
   if guid and group[guid] then
      return getRole(guid)
   end
end

function lib:getUnitSpecInfo(unit)
   local guid = unit and UnitGUID(unit)
   return lib:getGUIDSpecInfo(guid)
end

function lib:getGUIDSpecInfo(guid)
   local rec = guid and group[guid]
   if rec then
      return rec.specID, rec.specName, (rec.class or select(2, UnitClass(rec.unitID))), (getRole(guid)), rec.blizRole 
   end
end
